import rich
from rich.table import Table

from port_forward_manager.core import models


def prepare_sessions_table(show_link: bool = False):
    table = Table(show_edge=False, pad_edge=False, expand=True)

    table.add_column("ID", justify="right", style="yellow", width=12)
    table.add_column("Label", style="green", width=30)
    table.add_column("Hostname", justify="left", style="yellow", width=30)
    table.add_column("Type", justify="center", style="yellow", width=7)
    table.add_column("Local binding", justify="left", style="white", width=22)
    table.add_column("Remote binding", justify="left", style="blue", width=22)

    if show_link:
        table.add_column("URL", justify="left", style="cyan", width=60)

    return table


def prepare_schema_table(
    show_alias: bool = False,
    show_link: bool = False,
    show_schema: bool = True,
    title: str = "Schemas",
):

    table = Table(show_edge=False)

    table.add_column("ID", justify="right", style="green", width=12)
    table.add_column("Group", justify="left", style="yellow", width=20)
    table.add_column("Label", justify="left", style="yellow", width=30)
    table.add_column("Active", justify="right", style="yellow", width=30)
    return table


def prepare_group_table(
        show_alias: bool = False,
        show_link: bool = False,
        show_schema: bool = True,
        title: str = 'Schemas'
):

    table = Table(show_edge=False)

    table.add_column("ID", justify="right", style="green", width=12)
    table.add_column("Label", justify="left", style="yellow", width=30)

    return table


def prepare_ssh_group_table():

    table = Table(show_edge=False)

    table.add_column("ID", justify="right", style="green", width=12)
    table.add_column("Schema", justify="left", style="yellow", width=30)
    table.add_column("Label", justify="left", style="yellow", width=30)

    return table


def get_sessions_table(records: list) -> Table:
    table = prepare_sessions_table()
    for session in records:
        port = "auto" if session.local_port_dynamic else session.local_port.__str__()

        row = [
            session.id,
            session.label,
            session.hostname,
            session.type,
            f"{session.local_address}:{port}",
            f"{session.remote_address}:{session.remote_port.__str__()}",
        ]

        table.add_row(*row)

    if table.row_count > 0:
        table.rows[table.row_count - 1].end_section = True

    return table


def print_session(session: models.Session):
    table = prepare_sessions_table()
    row = [
        str(session.id),
        session.schema.name,
        session.hostname,
        session.type,
        session.local_address,
        "auto" if session.local_port_dynamic else session.local_port.__str__(),
        session.remote_address,
        session.remote_port.__str__(),
        session.label,
    ]

    table.add_row(*row)

    table.rows[table.row_count - 1].end_section = True
    rich.print(table)



from typing import List
from uuid import uuid4

from sqlmodel import Field, Relationship, SQLModel, select
from pydantic import computed_field

from port_forward_manager.core.database import get_session, engine

def generate_uid():
    return uuid4().hex[:12]


class Group(SQLModel, table=True):
    id: str = Field(default_factory=generate_uid, primary_key=True)
    label: str
    visible: bool = Field(default=True)
    order: int = Field(default=-1)

    schemas: list["Schema"] = Relationship(back_populates="group", cascade_delete=True)

    @staticmethod
    def delete(record: "Group"):
        db_session.delete(record)
        db_session.commit()

    @staticmethod
    def index(name=None):
        statement = select(Group).order_by(Group.label)

        if name:
            search = "%{}%".format(name)
            statement = statement.where(Group.label.like(search))

        cursor = db_session.exec(statement)
        return cursor.all()

    @staticmethod
    def find_by_id(group_id: str):
        statement = select(Group).where(Group.id == group_id)
        return db_session.exec(statement).first()

    @staticmethod
    def find_by_label(label):
        statement = select(Group).where(Group.label == label)
        return db_session.exec(statement).first()

    @staticmethod
    def add(group: "Group"):
        db_session.add(group)
        db_session.commit()

    @staticmethod
    def get_state():
        state = []
        for group in Group.index():
            state.append(group.dict())

        return state


class Schema(SQLModel, table=True):
    id: str = Field(default_factory=generate_uid, primary_key=True)
    label: str
    active: bool = Field(default=False)
    order: int = Field(default=-1)

    group_id: str = Field(default=None, foreign_key="group.id")
    group: Group = Relationship(back_populates="schemas")

    sessions: list['Session'] = Relationship(cascade_delete=True, back_populates="schema")
    ssh_groups: list['SSHGroup'] = Relationship(cascade_delete=True, back_populates="schema")

    @staticmethod
    def delete(schema: "Schema"):
        db_session.delete(schema)
        db_session.commit()

    @staticmethod
    def index(name=None, group_id=None):
        statement = select(Schema).order_by(Schema.label)

        if name:
            search = "%{}%".format(name)
            statement = statement.filter(Schema.label.like(search))

        if group_id:
            statement = statement.where(Schema.group_id == group_id)

        cursor = db_session.exec(statement)
        return cursor.all()

    @staticmethod
    def find_by_label(label: str):
        cursor = db_session.query(Schema)
        return cursor.filter(Schema.label == label).first()

    @staticmethod
    def find_by_id(schema_id: str):
        cursor = db_session.query(Schema)
        return cursor.filter(Schema.id == schema_id).first()

    def get_session(self, session_type: str, hostname: str, remote_port: int):
        for session in self.sessions:
            # print(session.dict())
            # print(hostname, remote_port)

            if (
                session.hostname == hostname
                and session.remote_port == remote_port
                and session_type == session.type
            ):
                # print(session.dict())
                return session

        # raise Exception(f"Session {hostname} and {remote_port} not found on {self.id}")

    def get_ssh_group(self, label: str):
        for group in self.ssh_groups:
            # print(session.dict())
            # print(hostname, remote_port)

            if group.label == label:
                # print(session.dict())
                return group

    @staticmethod
    def get_state():
        state = []
        for schema in Schema.index():
            schema_state = schema.dict()

            state.append(schema_state)

        return state


class SSHGroup(SQLModel, table=True):
    id: str = Field(default_factory=generate_uid, primary_key=True)
    label: str
    order: int = Field(default=-1)

    schema_id: str = Field(default=None, foreign_key="schema.id")
    schema: Schema = Relationship(back_populates="ssh_groups")

    @staticmethod
    def index(group_name=None) -> List["SSHGroup"]:
        cursor = db_session.query(SSHGroup)
        if group_name:
            search = "%{}%".format(group_name)
            cursor = cursor.filter(SSHGroup.label.like(search))

        return cursor.all()

    @staticmethod
    def find_by_id(group_id: str):
        cursor = db_session.query(SSHGroup)
        return cursor.filter(SSHGroup.id == group_id).first()

    @staticmethod
    def delete(group: "SSHGroup"):
        db_session.delete(group)
        db_session.commit()

    @staticmethod
    def get_state():
        state = []
        for group in SSHGroup.index():
            group_state = group.dict()

            state.append(group_state)

        return state


class Session(SQLModel, table=True):
    id: str = Field(default_factory=generate_uid, primary_key=True)
    label: str

    active: bool = Field(default=False)
    connected: bool = Field(default=False)

    hostname: str
    type: str = Field(default="local")
    remote_address: str = Field(default="127.0.0.1")
    remote_port: int = Field(default=0)
    local_address: str = Field(default="127.0.0.1")
    local_port: int = Field(default=0)
    local_port_dynamic: bool = Field(default=True)
    url_format: str = Field(default="http://{hostname}:{local_port}")
    auto_start: bool = Field(default=True)
    order: int = Field(default=-1)

    schema_id: str = Field(default=None, foreign_key="schema.id")
    schema: Schema = Relationship(back_populates="sessions")

    @staticmethod
    def clone(row):
        data = {}
        for column in row.__table__.columns:
            if column.name in ["id"]:
                continue

            data[column.name] = getattr(row, column.name)

        return Session(**data)

    @staticmethod
    def delete(session: "Session"):
        db_session.delete(session)
        db_session.commit()

    @staticmethod
    def find_by_id(session_id: str):
        statement = select(Session).where(Session.id == session_id)
        return db_session.exec(statement).first()

    @staticmethod
    def index(hostname=None, schema_id=None):
        statement = select(Session)

        if schema_id:
            statement = statement.where(Session.schema_id == schema_id)

        if hostname:
            search = "%{}%".format(hostname)
            statement = statement.filter(Session.hostname.like(search))

        cursor = db_session.exec(statement)
        return cursor.all()

    @computed_field(return_type=str)
    @property
    def cli_name(self) -> str:
        if self.schema and self.schema.group:
            return f"{self.schema.group.label}-{self.schema.label}".lower()
        return "Unnamed"

    @property
    def tmux_id(self) -> str:
        fields = [
            "pfm_session",
            "{schema_id}",
            "{hostname}",
            "{remote_address}",
            "{remote_port}",
            "{local_address}",
            "{local_port}",
            "{type}",
        ]

        data = {
            "schema_id": self.schema.id,
            "hostname": self.hostname,
            "remote_address": self.remote_address,
            "remote_port": self.remote_port,
            "local_address": self.local_address,
            "local_port": self.local_port,
            "type": self.type,
        }

        return "|".join(fields).format(**data).replace(".", "_")

    @property
    def command(self):
        ssh_options = [
            "-o ExitOnForwardFailure=yes",
            "-o ServerAliveCountMax=3",
            "-o ServerAliveInterval=10",
        ]

        session_data = self.dict()

        session_data["tmux_id"] = self.tmux_id
        session_data["shell_command"] = "ping localhost"
        session_data["options"] = " ".join(ssh_options)

        if self.type == "remote":
            ssh = "ssh {options} -R {local_address}:{local_port}:{remote_address}:{remote_port} {hostname}"
        else:
            ssh = "ssh {options} -L {local_address}:{local_port}:{remote_address}:{remote_port} {hostname}"

        session_data["ssh_command"] = ssh.format(**session_data)

        # start_command = "screen -dmS '{name}' {ssh_command}  -- {shell_command}"
        start_command = "tmux new-session -d -s '{tmux_id}' '{ssh_command} -- {shell_command}'".format(
            **session_data
        )
        # inspect(session_definition)
        return start_command

    @staticmethod
    def get_state():
        state = []
        for session in Session.index():
            session_state = session.dict()
            session_state["connected"] = session.connected
            # session_state["url"] = session.url
            state.append(session_state)
        return state


def init_database():
    print("Initialising database")
    SQLModel.metadata.create_all(engine)
    group = Group.find_by_label("port_forward_manager-system")
    if not group:
        print(f"Creating system group 'port_forward_manager-system'")
        group = Group(name="port_forward_manager-system", label="System", visible=False)
        schema = Schema(name="port_forward_manager-ephemeral", label="Ephemeral sessions")
        group.schemas.append(schema)
        Group.add(group)
        db_session.commit()


def reset_database():
    print("Resetting database")
    SQLModel.metadata.drop_all(engine)


db_session = get_session()



import subprocess
import random
from rich import print

from port_forward_manager.core.ui import prepare_sessions_table
from port_forward_manager.core import models, tools


def execute(command):
    # print(command)
    result = subprocess.run(
        command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
    )

    # inspect(result)
    # exit(1)
    if result.returncode != 0:
        print(f"Command [b]{command}[/] failed:\n {result.stderr}")
        exit(3)

    return result


def stop(session: models.Session):
    if session.connected:
        print("Stop session {tmux_id}.".format(tmux_id=session.tmux_id))
        stop_command = "tmux kill-session -t '{tmux_id}'".format(
            tmux_id=session.tmux_id
        )
        execute(stop_command)


def in_active_sessions(definition, active_sessions):
    for session in active_sessions:
        shared_items = {
            k: session[k]
            for k in session
            if k in definition and session[k] == definition[k]
        }
        # print(len(shared_items))
        if len(shared_items) >= 7:
            return True

    return False


def generate_new_port(port_list):
    while True:
        port = str(random.randint(9100, 9500))
        if port not in port_list:
            port_list.append(port)
            return port


def start(session, reconnect: bool = False):
    settings = tools.settings()
    if reconnect and session.connected:
        # print(f"[b]Stopping '{session_name}'[/b]")
        stop(session)

    if not session.connected:
        generate_port = False

        if not session.local_port:
            generate_port = True

        if session.local_port in settings.ports_active:
            if session.local_port_dynamic:
                print(
                    f"[b]Warning[/b]: Local port {session.local_port} already in use, assigning new port"
                )
                generate_port = True
            else:
                print(f"[b]Warning[/b]: Local port {session.local_port} already in use")
                exit()

        if generate_port:
            session.local_port = generate_new_port(settings.ports_active)

        result = execute(session.command)
        return result
    else:
        print(
            "Ignoring {type} forward on [b]{hostname}[/b] {remote_address}:{remote_port}.".format(
                **session.as_dict()
            )
        )


def refresh_state():
    update_state()
    for session in models.Session.index():
        if not session.active or session.connected:
            continue

        start(session)


def update_state():
    settings = tools.settings()

    # Reset state
    settings.ports_active = []
    for schema in models.Schema.index():
        schema.active = False

        for session in schema.sessions:
            session.connected = False

    # Update state from tmux sessions
    command = "tmux ls | grep 'pfm_session' | cut -d ':' -f 1"
    result = execute(command)

    session_id_list = result.stdout.split("\n")

    for session_id in session_id_list:
        if len(session_id) == 0:
            continue

        session_data = session_id.replace("_", ".").replace(":", "")
        # print(f"Active -> {session_id}")
        values = session_data.split("|")

        if len(values) == 8:
            (
                filler,
                schema_id,
                hostname,
                remote_host,
                remote_port,
                local_address,
                local_port,
                forward_type,
            ) = values

            schema = models.Schema.find_by_id(schema_id)
            if schema:
                # schema.active = True
                session = schema.get_session(forward_type, hostname, int(remote_port))
                if not session:
                    # print(f"missing session definition {hostname} {remote_port} {forward_type}")
                    continue

                session.connected = True
                schema.active = True
                # session.local_address = local_address
                session.local_port = local_port
                settings.ports_active.append(local_port)
            else:
                continue
                # print("TMUX session active but missing session definition.")


def filter_session(
    session: models.Session, schema_id: str = None, host: str = None, port: str = None
):
    if schema_id and schema_id != session.schema_id:
        return True
    if host and host not in session.hostname:
        return True
    if port and port != session.remote_port:
        return True


def list_from_active(key, filter_string: str = ""):
    items = []

    for schema in models.Schema.index():
        for session in schema.sessions:
            if not session.active:
                continue

            item = session.label
            if filter_string not in item:
                continue

            if item not in items:
                items.append(item)

    return items


def show_active_sessions(schema: str = None, host: str = None, port: int = None):
    update_state()

    table = prepare_sessions_table(True)
    session_count = 0
    for schema in models.Schema.index():
        schema_rows = False
        for session in schema.sessions:
            if not session.connected:
                continue

            schema_rows = True

            row = [
                f"{schema.label}",
                session.hostname,
                session.type,
                session.local_address,
                "auto" if session.local_port_dynamic else session.local_port.__str__(),
                session.remote_address,
                session.remote_port.__str__(),
                session.url,
            ]

            table.add_row(*row)
            session_count += 1

        if schema_rows:
            table.rows[table.row_count - 1].end_section = True

    if session_count > 0:
        print(table)
    else:
        print("Nothing to see here")



import rich

from port_forward_manager.core import models


def base_normalisation(record):
    record_id = str(record.get("id"))
    if len(str(record_id)) != 12:
        record["id"] = models.generate_uid()

    return record_id


def normalise_import(settings):
    rich.print("Normalise import")

    group_index = {}
    schema_index = {}
    session_index = {}
    ssh_group_index = {}

    for record in settings.get("groups", []):
        record_id = base_normalisation(record)

        group_index[record_id] = models.Group(**record)

    # rich.print("Groups", group_index)

    for record in settings.get("schemas", []):
        record_id = base_normalisation(record)

        previous_group_id = str(record.get("group_id"))
        group = group_index.get(previous_group_id)
        if not group:
            rich.print(f"Error: Group not found {previous_group_id}")
            continue

        record['group_id'] = group.id
        record["active"] = False

        schema_index[record_id] = models.Schema(**record)

    # rich.print("Schemas", schema_index)

    for record in settings.get("sessions", []):
        record_id = base_normalisation(record)

        schema_id = str(record.get("schema_id"))
        schema = schema_index.get(schema_id)

        if not schema:
            rich.print(f"Error: Session schema not found {schema_id}")
            continue

        record["schema_id"] = schema.id

        session_index[record_id] = models.Session(**record)

    # rich.print("Sessions", session_index)

    for record in settings.get("ssh_groups", []):
        record_id = str(record.get("id"))

        if len(str(record_id)) == 12:
            record["id"] = models.generate_uid()

        schema_id = str(record.get("schema_id"))
        schema = schema_index.get(schema_id)
        if not schema:
            rich.print(f"Error: SSHGroup schema not found {schema_id}")
            continue

        record["schema_id"] = schema.id
        ssh_group_index[record_id] = models.SSHGroup(**record)

    # rich.print("SSHGroups", ssh_group_index)

    settings["groups"] = list(group_index.values())
    settings["schemas"] = list(schema_index.values())
    settings["sessions"] = list(session_index.values())

    return settings


import typer
import time
from enum import Enum
import rich
from rich.prompt import Confirm

from port_forward_manager.core.autocomplete import ac_schemas
from port_forward_manager.core.ui import get_sessions_table, print_session
from port_forward_manager.core import models, tools, database
from port_forward_manager.core.forward_sessions import update_state, start as start_session, stop as stop_session

app = typer.Typer(no_args_is_help=True)
db = database.get_session()


class SessionTypes(str, Enum):
    local = "local"
    remote = "remote"
    ssh = "ssh"


@app.command()
def index(
    hostname: str = typer.Argument(None, help="Hostname"),
    active: bool = typer.Option(False, help="Active sessions"),
    schema: str = typer.Option(None, help="Filter by schema", autocompletion=ac_schemas),
):
    """
    List sessions
    """
    records = []
    for session in models.Session.index(hostname, schema_id=schema):
        if active and not session.active:
            continue

        records.append(session)

    update_state()
    table = get_sessions_table(records)
    rich.print(table)


@app.command()
def create(
    schema_name: str = typer.Argument(None, help="Schema name"),
    session_type: SessionTypes = typer.Argument(
        SessionTypes.local, help="Session type", case_sensitive=False
    ),
    label: str = typer.Argument(None, help="Session label"),
    hostname: str = typer.Argument(None, help="Session hostname"),
    remote_address: str = typer.Option("127.0.0.1", help="Remote address"),
    remote_port: int = typer.Option(0, help="Remote address", min=0, max=65535),
    local_address: str = typer.Option("127.0.0.1", help="Remote address"),
    local_port: int = typer.Option(0, help="Remote address", min=0, max=65535),
    local_port_dynamic: bool = typer.Option(True, help="Assign local port dynamically"),
    auto_start: bool = typer.Option(True, help="Remote address"),
    url_format: str = typer.Option(
        "http://{hostname}:{local_port}", help="Dynamic URL"
    ),
    schema_id: bool = typer.Option(True, help="Use schema ID instead of name"),
):
    """
    Create session
    """
    if schema_id:
        schema = models.Schema.find_by_id(int(schema_name))
    else:
        schema = models.Schema.find_by_label(schema_name)

    if not schema:
        print(f"Schema '{schema_name}' doesn't exist")
        exit(401)

    session = models.Session()
    session.type = session_type.value
    session.label = label
    session.hostname = hostname
    session.remote_address = remote_address
    session.remote_port = remote_port
    session.local_address = local_address
    session.local_port = local_port
    session.local_port_dynamic = local_port_dynamic
    session.auto_start = auto_start
    session.url_format = url_format
    schema.sessions.append(session)
    models.db_session.commit()
    print_session(session)


@app.command()
def delete(
    session_id: str = typer.Argument(None, help="Session ID"),
    force: bool = typer.Option(False, help="Force delete"),
):
    """Delete session"""
    session = models.Session.find_by_id(session_id)
    if force or Confirm.ask(f"Are you sure you want to delete '{session.label}'?"):
        models.Session.delete(session)


@app.command()
def update(
    session_id: str = typer.Argument(None, help="Session ID"),
    schema_name: str = typer.Option(None, help="Schema name"),
    session_type: SessionTypes = typer.Option(
        None, help="Session type", case_sensitive=False
    ),
    label: str = typer.Option(None, help="Session label"),
    hostname: str = typer.Option(None, help="Session hostname"),
    remote_address: str = typer.Option(None, help="Remote address"),
    remote_port: int = typer.Option(None, help="Remote address", min=0, max=65535),
    local_address: str = typer.Option(None, help="Remote address"),
    local_port: int = typer.Option(None, help="Remote address", min=0, max=65535),
    local_port_dynamic: bool = typer.Option(None, help="Assign local port dynamically"),
    auto_start: bool = typer.Option(None, help="Remote address"),
    url_format: str = typer.Option(None, help="Dynamic URL"),
    schema_id: bool = typer.Option(True, help="Use schema ID instead of name"),
):
    """
    Update session settings
    """

    session = models.Session.find_by_id(session_id)
    if not session:
        print("Session ID not found")
        exit(401)

    if session_type:
        session.type = session_type.value

    if label:
        session.label = label

    if hostname:
        session.hostname = hostname

    if remote_address:
        session.remote_address = remote_address

    if remote_port:
        session.remote_port = remote_port

    if local_address:
        session.local_address = local_address

    if local_port:
        session.local_port = local_port

    if local_port_dynamic:
        session.local_port_dynamic = local_port_dynamic

    if auto_start:
        session.auto_start = auto_start

    if url_format:
        session.url_format = url_format

    if schema_name:
        if schema_id:
            schema = models.Schema.find_by_id(int(schema_name))
        else:
            schema = models.Schema.find_by_label(schema_name)
        if schema is None:
            print(f"Schema '{schema_name}' doesn't exist")
            exit(401)

        session.schema_id = schema.id

    models.db_session.commit()
    print_session(session)


@app.command()
def connect(
    label: str = typer.Argument(..., help="Session label"),
    hostname: str = typer.Argument(..., help="Session hostname"),
    remote_port: int = typer.Argument(..., help="Remote address", min=0, max=65535),
    session_type: SessionTypes = typer.Option(
        SessionTypes.local, help="Session type", case_sensitive=False
    ),
    remote_address: str = typer.Option("127.0.0.1", help="Remote address"),
    local_address: str = typer.Option("127.0.0.1", help="Remote address"),
    local_port: int = typer.Option(0, help="Remote address", min=0, max=65535),
):
    """
    Connect ephemeral session
    """
    settings = tools.settings()

    schema_name = "port_forward_manager-ephemeral"
    schema = models.Schema.find_by_label(schema_name)

    if schema is None:
        print(f"Schema '{schema_name}' doesn't exist")
        exit(401)

    session = models.Session()
    session.type = session_type.value
    session.label = label
    session.hostname = hostname
    session.remote_address = remote_address
    session.remote_port = remote_port
    session.local_address = local_address
    session.local_port = local_port
    session.local_port_dynamic = local_port == 0
    session.auto_start = False
    session.url_format = ""
    schema.sessions.append(session)
    models.db_session.commit()

    start(session, False)
    session.active = True
    session.connected = True

    time.sleep(settings.wait_after_start)
    models.db_session.commit()

    print_session(session)


@app.command()
def start(session_id: str = typer.Argument(None, help="Session ID")):
    """
    Start session
    """
    update_state()
    session = models.Session.find_by_id(session_id)

    if not session:
        print("Session doesn't exist")
        exit()

    session.active = True

    start_session(session)
    models.db_session.commit()


@app.command()
def stop(session_id: str = typer.Argument(None, help="Session ID")):
    """
    Disconnect session
    """
    update_state()
    session = models.Session.find_by_id(session_id)

    if not session:
        print("Session doesn't exist")
        exit()

    session.active = False

    stop_session(session)

    if session.schema.name == "port_forward_manager-ephemeral":
        print("Delete ADHOC session")
        models.Session.delete(session)

    models.db_session.commit()


@app.command()
def disconnect(session_id: str = typer.Argument(None, help="Session ID")):
    """
    Disconnect session
    """
    update_state()
    session = models.Session.find_by_id(session_id)

    if not session:
        print("Session doesn't exist")
        exit()

    if Confirm.ask(f"Are you sure you want to disconnect '{session.label}'?"):
        stop(session)

        if session.schema.name == "port_forward_manager-ephemeral":
            print("Delete ADHOC session")
            models.Session.delete(session)



from importlib.metadata import version as current_version
import os.path
import rich
from rich.prompt import Confirm
from sshconf import read_ssh_config
import time
import typer
import yaml

from port_forward_manager.core import models, tools, forward_sessions
from port_forward_manager.core import autocomplete
from port_forward_manager import cli_group, cli_schema, cli_session, cli_ssh_group
from port_forward_manager.core.normalisation import normalise_import

app = typer.Typer(no_args_is_help=True, pretty_exceptions_enable=False)
app.add_typer(cli_group.app, name="group", help="Group management")
app.add_typer(cli_schema.app, name="schema", help="Schema management")
app.add_typer(cli_session.app, name="session", help="Session management")
app.add_typer(cli_ssh_group.app, name="ssh_group", help="SSH group management")
tools.load_settings()


@app.command()
def shutdown():
    """
    Stop all active sessions
    """

    settings = tools.settings()

    forward_sessions.update_state()
    for schema in models.Schema.index():
        for session in schema.sessions:
            if session.connected:
                forward_sessions.stop(session)
            session.active = False
        schema.active = False

    time.sleep(settings.wait_after_stop)
    forward_sessions.show_active_sessions()
    models.db_session.commit()


@app.command()
def status(
    schema: str = typer.Option(None, shell_complete=autocomplete.sc_schemas),
    host: str = typer.Option(None, shell_complete=autocomplete.sc_hosts),
    port: int = None,
    json: bool = typer.Option(False, "--json", "-j", help="Output JSON"),
):
    """
    Show active sessions
    """

    forward_sessions.show_active_sessions(schema, host, port)


@app.command()
def state():
    """
    Show current state in JSON format
    """

    ssh_config = read_ssh_config(os.path.expanduser("~/.ssh/config"))
    ssh_hosts = []
    for host in ssh_config.hosts():
        if "*" in host:
            continue
        hosts = host.split(" ")
        for hostname in hosts:
            ssh_hosts.append(hostname)

    forward_sessions.refresh_state()
    time.sleep(0.5)
    forward_sessions.update_state()
    current_state = {
        "groups": models.Group.get_state(),
        "schemas": models.Schema.get_state(),
        "sessions": models.Session.get_state(),
        "ssh_groups": models.SSHGroup.get_state(),
        "ssh_hosts": ssh_hosts,
    }

    print(yaml.dump(current_state))
    # print(simplejson.dumps(current_state, indent=2))


@app.command()
def version():
    """
    Show PFM version
    """
    version_string = current_version("port_forward_manager")
    rich.print(f"Port Forward Manager [bold white]v{version_string}[/]")


@app.command()
def db_wipe():
    """Wipe the whole database clean"""
    if Confirm.ask(
        "This action will wipe all groups, schemas and sessions, are you sure?"
    ):
        models.reset_database()
        models.init_database()


@app.command()
def yaml_export(
    export_path: str = typer.Argument(None, help="YAML configuration file")
):
    """Import groups, schemas and sessions from configuration file"""
    export_data = {
        "export_format": "db_dump",
        "version": "2.0",
        "groups": models.Group.get_state(),
        "schemas": models.Schema.get_state(),
        "sessions": models.Session.get_state(),
        "ssh_groups": models.SSHGroup.get_state(),
    }

    yaml_string = yaml.dump(export_data)
    if export_path:
        tools.write_file(export_path, yaml_string)
    else:
        print(yaml_string)


@app.command()
def yaml_import(
    export_path: str = typer.Argument(..., help="YAML configuration file"),
    wipe: bool = typer.Option(False, help="Wipe DB"),
    force: bool = typer.Option(False, help="Wipe DB"),
):
    """Import groups, schemas and sessions from configuration file"""
    settings = tools.load_yaml_file(export_path)

    change_count = 0

    if settings.get("export_format") != "db_dump":
        rich.print("[red]Invalid export file format[/]")
        exit()

    if wipe and not force:
        if Confirm.ask(
            "This action will wipe all groups, schemas and sessions, are you sure?"
        ):
            models.reset_database()
            models.init_database()
    elif wipe and force:
        models.reset_database()
        models.init_database()

    settings = normalise_import(settings)

    rich.print("Importing PFM DB DUMP")
    groups = {}
    schemas = {}
    sessions = {}

    # Import groups
    for group_definition in settings.get("groups", []):
        group = models.Group.find_by_id(group_definition.id)
        if not group:
            group = models.Group.find_by_label(group_definition.label)

        if not group:
            change_count += 1
            rich.print(f"* Creating group {group_definition.label} {group_definition.id}")
            models.Group.add(group_definition)
        else:
            for key, value in group_definition.dict().items():
                if key == "id":
                    group_definition.id = group.id
                    continue
                setattr(group, key, value)

        groups[group_definition.id] = group_definition

    # rich.print(f"Group insert: { groups }")

    for group in models.Group.index():
        # rich.print(f"* Checking group {group.label}")
        if group.id not in groups.keys():
            if not group.visible:
                continue

            rich.print(f"  * Deleting group {group.label} - {group.id}")
            models.Group.delete(group)

    rich.print(f"Imported {change_count} groups")
    models.db_session.commit()

    # Import schemas
    for schema_definition in settings.get('schemas', []):
        tmp_group = groups.get(schema_definition.group_id, {})

        # rich.print(f"  * Group", schema_definition.group_id, groups)
        group = models.Group.find_by_id(schema_definition.group_id)

        if not group:
            group = models.Group.find_by_label(tmp_group.get('label'))

        if not group:
            rich.print(f"IGNORING schema - Error could not find group {tmp_group.get('label')}")
            continue

        schemas[schema_definition.id] = schema_definition.label

        schema = models.Schema.find_by_label(schema_definition.label)
        if not schema:
            change_count += 1
            rich.print(f"* Importing schema {schema_definition.label}")
            group.schemas.append(schema_definition)
        elif not schema.active:
            for key, value in schema_definition.dict().items():
                setattr(schema, key, value)

    for schema in models.Schema.index():
        # rich.print(f"* Checking schema {schema.label}")
        if schema.id not in schemas.keys() and not schema.active:
            rich.print(f"  * Deleting group")
            models.Schema.delete(schema)

    models.db_session.commit()

    # Import sessions
    for session_definition in settings.get("sessions"):
        schema_label = schemas.get(session_definition.schema_id)
        rich.print(f"(yaml-import) Schema: {schema_label} '{session_definition.schema_id}'")

        schema = models.Schema.find_by_id(session_definition.schema_id)
        if not schema:
            rich.print(
                f"IGNORING session - Error could not find schema '{schema_label}'"
            )
            continue

        session = schema.get_session(
            session_definition.type,
            session_definition.hostname,
            session_definition.remote_port
        )

        if not session:
            session = models.Session.find_by_id(session_definition.id)

        if not session:
            change_count += 1
            print(f"    * Importing session {session_definition.type} {session_definition.hostname} {session_definition.remote_port}")
            if session_definition.local_port == 0:
                session_definition.local_port_dynamic = True

            schema.sessions.append(session_definition)
        else:
            for key in session.__fields__.keys() - ['connected', 'active']:
                value = getattr(session_definition, key)
                setattr(session, key, value)

        sessions[session_definition.id] = session_definition

    for session in models.Session.index():
        # rich.print(f"* Checking session {session.label}")
        if session.id not in sessions.keys() and not session.connected:
            rich.print(f"  * Deleting session")
            models.Session.delete(session)

    models.db_session.commit()

    # Import SSH groups
    for ssh_group in settings.get("sshGroups", []):
        schema = models.Schema.find_by_id(ssh_group.schema_id)
        if not schema:
            rich.print(
                f"IGNORING session - Error could not find schema '{ssh_group.schema_id}'"
            )
            continue

        group = schema.get_ssh_group(ssh_group.label)
        if not group:
            change_count += 1
            print(f"    * Importing SSH group {ssh_group.label}")

            schema.ssh_groups.append(ssh_group)

    models.db_session.commit()

    if change_count == 0:
        rich.print("No changes...")
    else:
        rich.print("There were {} items imported.".format(change_count))


def run():
    #models.init_database()
    app()




import time
import rich
from rich.prompt import Confirm
import typer

from port_forward_manager.core import models, tools, database, forward_sessions
from port_forward_manager.core import autocomplete
from port_forward_manager.core.ui import prepare_sessions_table, prepare_schema_table
from port_forward_manager.core.forward_sessions import show_active_sessions

app = typer.Typer(no_args_is_help=True)
db = database.get_session()


@app.command()
def create(
    group_id: str = typer.Argument(..., help="Group name"),
    label: str = typer.Argument(..., help="Schema label"),
):
    """
    Create a schema
    """
    group = models.Group.find_by_id(group_id)
    if group is None:
        print(f"Group '{group_id}' doesn't exist")
        exit(401)

    schema = models.Schema(label=label, group_id=group.id)
    group.schemas.append(schema)

    models.db_session.commit()


@app.command()
def update(
    schema_id: str = typer.Argument(..., help="Schema ID"),
    group_id: str = typer.Option(None, help="Group ID"),
    label: str = typer.Option(None, help="Schema label"),
):
    """
    Create a schema
    """
    schema = models.Schema.find_by_id(schema_id)
    if schema is None:
        print(f"Schema '{schema_id}' doesn't exist")
        exit(401)

    if group_id:
        group = models.Group.find_by_id(group_id)
        if not group:
            print("Group is invalid")
        else:
            schema.group_id = group_id

    if label:
        schema.label = label

    models.db_session.commit()


@app.command()
def index(
        schema_filter: str = typer.Argument(None, shell_complete=autocomplete.sc_schemas),
        gid: str = typer.Option(None, help="Group ID", autocompletion=autocomplete.ac_groups)
    ):
    """
    List configured schemas
    """
    settings = tools.settings()
    table = prepare_schema_table(True, settings.show_schema_link, True)

    print(f"Group: {gid}")

    for schema in models.Schema.index(schema_filter, gid):
        if schema.group_id is None:
            print(f"Schema #{schema.id} group is INVALID")
            row = [
                str(schema.id),
                "INVALID GROUP",
                schema.label,
                "Yes" if schema.active else "No",
            ]
        else:
            row = [
                str(schema.id),
                schema.group.label,
                schema.label,
                "Yes" if schema.active else "No",
            ]
        table.add_row(*row)

    if table.row_count > 0:
        table.rows[table.row_count - 1].end_section = True

    rich.print(table)


@app.command()
def clone(
    schema_id: str = typer.Argument(..., help="Schema ID"),
    group_id: str = typer.Option(None, help="Group ID"),
    label: str = typer.Option(None, help="Schema label"),
):

    schema_original = models.Schema.find_by_id(schema_id)
    if schema_original is None:
        print(f"Schema {schema_id} not found")
        exit(3)

    if group_id is None:
        group_id = schema_original.group_id

    group = models.Group.find_by_id(group_id)
    if not group:
        print(f"Group {group_id} not found")
        exit(3)

    if not label:
        label = f"{schema_original.label} clone"

    schema_clone = models.Schema(label=label)
    group.schemas.append(schema_clone)

    for session_original in schema_original.sessions:
        session_clone = models.Session.clone(session_original)

        print(f"Clone session {session_clone.label}")
        if session_clone.local_port_dynamic:
            session_clone.local_port = 0

        schema_clone.sessions.append(session_clone)

    models.db_session.commit()


@app.command()
def delete(
    schema_id: str = typer.Argument(None, help="Schema ID"),
    force: bool = typer.Option(False, help="Force delete"),
):
    """Delete session"""

    schema = models.Schema.find_by_id(schema_id)
    if schema is None:
        print("Schema not found")
        exit()

    if force or Confirm.ask(f"Are you sure you want to delete '{schema.label}'?"):
        models.Schema.delete(schema)


@app.command()
def sessions(schema_filter: str = typer.Argument(None, autocompletion=autocomplete.ac_schemas)):
    """
    List configured schemas
    """
    table = prepare_sessions_table()

    for schema in models.Schema.index(schema_filter):
        for session in schema.sessions:
            row = [
                str(schema.id),
                session.hostname,
                session.type,
                session.local_address,
                "-----" if session.local_port_dynamic else session.local_port.__str__(),
                session.remote_address,
                session.remote_port.__str__(),
                session.label,
            ]
            table.add_row(*row)
        if table.row_count > 0:
            table.rows[table.row_count - 1].end_section = True

    rich.print(table)


@app.command()
def start(
    schema_id: str = typer.Argument(..., autocompletion=autocomplete.ac_schemas),
    force: bool = typer.Option(None, help="Force sessions reconnection"),
):
    """
    Start a schema of forwarding sessions
    """
    settings = tools.settings()
    forward_sessions.update_state()

    schema = models.Schema.find_by_id(schema_id)

    if schema is None:
        print("[b]Schema '{0}' is unknown[/b]".format(schema_id))
        exit(-1)

    schema.active = True

    for session in schema.sessions:
        session.schema_id = schema.id
        if session.auto_start:
            forward_sessions.start(session, force)
            session.active = True

    time.sleep(settings.wait_after_start)
    show_active_sessions()
    models.db_session.commit()


@app.command()
def stop(
    schema_id: str = typer.Argument(None, autocompletion=autocomplete.ac_active_schemas),
    hostname: str = typer.Option(None, shell_complete=autocomplete.sc_active_hosts),
    port: str = typer.Option(None, shell_complete=autocomplete.sc_active_remote_port),
):
    """
    Stop sessions from active schema, host or port
    """

    if not schema_id and not hostname and not port:
        print("[b]Pick a schema, host or host and port or --all[/b]")
        exit(-1)

    settings = tools.settings()
    forward_sessions.update_state()

    if schema_id:
        schema = models.Schema.find_by_id(schema_id)

        if not schema:
            print("[b]Schema not found")
            exit(-1)

        for session in schema.sessions:

            forward_sessions.stop(session)
            session.active = False
        schema.active = False

    time.sleep(settings.wait_after_stop)
    show_active_sessions()
    models.db_session.commit()



from click.shell_completion import CompletionItem, Context
from sshconf import read_ssh_config
from os.path import expanduser

from port_forward_manager.core import models, forward_sessions


# Auto completion
def ac_groups(incomplete):
    result = []

    for group in models.Group.index():
        if incomplete not in group.label:
            continue

        item = (group.id, f"{len(group.schemas)} schema(s) for {group.label}")
        result.append(item)
    return result


def ac_schemas(incomplete):
    schemas = models.Schema.index()
    result = []

    for schema in schemas:
        if not schema.label.startswith(incomplete):
            continue

        item = (schema.id, f"{len(schema.sessions)} session(s) for {schema.group.label} - {schema.label}")
        result.append(item)
    return result


def ac_active_schemas(incomplete):
    items = []
    for schema in models.Schema.index():
        if schema.active:
            items.append(schema.name)

    return items


def ac_hosts(incomplete):
    result = []
    c = read_ssh_config(expanduser("~/.ssh/config"))
    for host in c.hosts():
        if "*" not in host and host.startswith(incomplete):
            hosts = host.split(" ")
            for hostname in hosts:
                result.append(hostname)

    return result


# Shell completion
def sc_schemas(ctx: Context, param, incomplete):
    result = []

    group_id = ctx.params.get("gf")
    schemas = models.Schema.index(group_id=group_id)

    item = CompletionItem("Hello", help=f"{ctx.params}")
    result.append(item)
    for schema in schemas:
        if not group_id and incomplete not in schema.label:
            continue

        item = CompletionItem(schema.label, help=f"{len(schema.sessions)} session(s) for {schema.label} {ctx.params}")
        result.append(item)

    return result


def sc_active_schemas(ctx, param, incomplete):
    return forward_sessions.list_from_active('schema', incomplete)


def sc_active_remote_port(ctx, param, incomplete):
    return forward_sessions.list_from_active('remote_port')


def sc_active_hosts(ctx, param, incomplete):
    return forward_sessions.list_from_active('hostname')


def sc_hosts(ctx, param, incomplete):
    from sshconf import read_ssh_config
    from os.path import expanduser
    result = []
    c = read_ssh_config(expanduser("~/.ssh/config"))
    for host in c.hosts():
        if "*" not in host and host.startswith(incomplete):
            hosts = host.split(" ")
            for hostname in hosts:
                item = CompletionItem(hostname)
                result.append(item)

    return result



import os
from sqlmodel import create_engine, Session, SQLModel

from port_forward_manager.core import tools

database_path = os.path.join(tools.base_path, "database.db")
database_url = f"sqlite:///{database_path}"
engine = create_engine(database_url)


def get_session():
    return Session(engine)
